﻿using System;
using System.Collections.Generic;

// Also see:
// https://github.com/dotnet/corefx/blob/master/src/System.Linq/src/System/Linq/Enumerable.cs

namespace AdrianKousz.Util
{
	public static class Enumerables
	{
		#region Fillers

		public static T[] Fill<T>(this T[] array) where T : new() { return Fill(array, i => new T()); }
		public static T[] Fill<T>(this T[] array, T value) { return Fill(array, i => value); }
		public static T[] Fill<T>(this T[] array, Func<T> f) { return Fill(array, i => f()); }
		public static T[] Fill<T>(this T[] array, Func<int, T> f)
		{
			var min = array.GetLowerBound(0);
			var max = array.GetUpperBound(0);
			for (var i = min; i <= max; i++)
				array[i] = f(i);
			return array;
		}

		public static IList<T> Fill<T>(this IList<T> list, int count) where T : new() { return Fill(list, i => new T(), count); }
		public static IList<T> Fill<T>(this IList<T> list, T value, int count) { return Fill(list, i => value, count); }
		public static IList<T> Fill<T>(this IList<T> list, Func<T> f, int count) { return Fill(list, i => f(), count); }
		public static IList<T> Fill<T>(this IList<T> list, Func<int, T> f, int count)
		{
			for (var i = 0; i < count; i++)
				list.Add(f(i));
			return list;
		}

		#endregion

		#region Functional Classics

		public static int IndexOf<T>(this IEnumerable<T> source, T value, Func<T, T, bool> comparer)
		{
			var i = 0;
			foreach (var el in source) {
				if (comparer(el, value)) return i;
				i++;
			}
			return -1;
		}

		public static IEnumerable<T> ForEach<T>(this IEnumerable<T> source, Action<T> f) { return ForEach(source, (i, x) => f(x)); }
		public static IEnumerable<T> ForEach<T>(this IEnumerable<T> source, Action<int, T> f)
		{
			var i = 0;
			foreach (var el in source) f(i++, el);
			return source;
		}

		// Linq: Select
		public static IEnumerable<TResult> Map<T, TResult>(this IEnumerable<T> source, Func<T, TResult> f)
		{
			foreach (var el in source) yield return f(el);
		}

		// Linq: Where
		public static IEnumerable<T> Filter<T>(this IEnumerable<T> source, Func<T, bool> f)
		{
			foreach (var el in source) if (f(el)) yield return el;
		}

		public static TResult FoldL<T, TResult>(this IEnumerable<T> source, TResult initial, Action<TResult, T> f)
		{ return FoldL(source, initial, (folded, el) => { f(folded, el); return folded; }); }
		public static TResult FoldL<T, TResult>(this IEnumerable<T> source, TResult initial, Func<TResult, T, TResult> f)
		{
			var folded = initial;
			foreach (var el in source) folded = f(folded, el);
			return folded;
		}

		// Linq: Aggregate
		public static T FoldL1<T>(this IEnumerable<T> source, Func<T, T, T> f)
		{
			using (var en = source.GetEnumerator()) {
				if (!en.MoveNext()) throw new InvalidOperationException("No elements");
				T folded = en.Current;
				while (en.MoveNext()) folded = f(folded, en.Current);
				return folded;
			}
		}

		public static IEnumerable<TResult> Zip<T1, T2, TResult>(this IEnumerable<T1> s1, IEnumerable<T2> s2, Func<T1, T2, TResult> f)
		{
			using (var en1 = s1.GetEnumerator()) {
				using (var en2 = s2.GetEnumerator()) {
					while (en1.MoveNext() && en2.MoveNext()) {
						yield return f(en1.Current, en2.Current);
					}
				}
			}
		}

		#endregion
	}
}

